Unit SmModeX;

{ Lite version of ModeX unit, not as much of a memory hog }

Interface

Const
  SCREEN_SEG            = $A000; { Segment of display memory in mode X      }
  SCREEN_WIDTH          = 320;
  SCREEN_HEIGHT         = 240;
  PAGE0_START_OFFSET    = 0;
  PAGE1_START_OFFSET    = (SCREEN_HEIGHT * SCREEN_WIDTH) Div 4;
  BG_START_OFFSET       = (SCREEN_HEIGHT * SCREEN_WIDTH * 2) Div 4;
  DOWNLOAD_START_OFFSET = (SCREEN_HEIGHT * SCREEN_WIDTH * 3) Div 4;
  SC_INDEX       = $03C4;     { Sequence Controller Index                   }
  CRTC_INDEX     = $03D4;     { CRT Controller Index                        }
  MISC_OUTPUT    = $03C2;     { Miscellaneous Output Register               }
  MAP_MASK       = 2;         { Index in SC of Map Mask register            }
  GC_INDEX       = $03CE;     { Graphics Controller Index register port     }
  BIT_MASK       = 8;         { Index in GC of Bit Mask register            }
  READ_MAP       = 4;         { Index in GC of Read Map Select register     }

  _32BIT         = $66;       { Used for 32-bit instructions; overrides the }
                              {  operand size attributes for the following  }
                              {  instruction to force it to 32 bits.        }
                              {                                             }
                              { Example:                                    }
                              {                                             }
                              {    DB    _32BIT                             }
                              {    MOV   AX,WORD PTR I                      }
                              {                                             }
                              { Really looks like when run:                 }
                              {                                             }
                              {    MOV   EAX,DWORD PTR I                    }
                              {                                             }
                              { This allows 32-bit instructions to be used  }
                              {  in BASM, without using "InLine"s.          }

Procedure FillRectangleX(StartX,StartY,EndX,EndY: Integer; PageBase: Word; Color: Integer);
Procedure DrawLineX(X0,Y0,X1,Y1: Integer; PageBase: Word; Color: Word);
Procedure RectangleX(StartX,StartY,EndX,EndY: Integer; PageBase: Word; Color: Word);
Procedure CopySystemToScreenX(SourceStartX,SourceStartY,SourceEndX,SourceEndY,
  DestStartX,DestStartY: Integer; SourcePtr: Pointer; DestPageBase: Word;
  SourceBitmapWidth,DestBitmapWidth: Integer);

Implementation

Const
  _SCREEN_WIDTH  = SCREEN_WIDTH Div 4; { Width of screen in bytes from one  }
                                       {  scan line to the next             }
  PATTERN_BUFFER = $FFFC;     { Offset in screen memory of the buffer used  }
                              {  to store each pattern during drawing       }
  INPUT_STATUS_1 =  $03DA;    { Input Status 1 register                     }
  START_ADDRESS_HIGH = $0C;   { Bitmap start address high byte              }
  START_ADDRESS_LOW  = $0D;   { Bitmap start address low byte               }

  { Plane masks for clipping left and right edges of rectangle. }

  LeftClipPlaneMask  : Array[0..3] of Byte = ($0F,$0E,$0C,$08);
  RightClipPlaneMask : Array[0..3] Of Byte = ($0F,$01,$03,$07);

Procedure FillRectangleX(StartX,StartY,EndX,EndY: Integer; PageBase: Word; Color: Integer);
{ ------------------------------------------------------------------------- }
{ Mode X (320x240, 256 colors) rectangle fill routine. Works on all VGAs.   }
{ Uses fast approach that fans data out to up to four planes at once to     }
{ draw up to four pixels at once. Fills up to but not including the column  }
{ atr EndX and the row at EndY. No clipping is performed.                   }
{ Pascal near-callable as:                                                  }
{    FillRectangleX(StartX,StartY,EndX,EndY,PageBase,Color);                }
{ ------------------------------------------------------------------------- }

Begin
  Asm
    CLI
    PUSH   SI               { Preserve caller's register variables          }
    PUSH   DI

    CLD
    MOV    AX,_SCREEN_WIDTH
    MUL    WORD PTR StartY  { Offset in page of top rectangle scan line     }
    MOV    DI,WORD PTR StartX
    SHR    DI,1             { X/4 = Offset of first rectangle pixel in scan }
    SHR    DI,1             {  line                                         }
    ADD    DI,AX            { Offset of first rectangle pixel in page       }
    ADD    DI,WORD PTR PageBase { Offset of first rectangle pixel in display}
                            {  memory                                       }
    MOV    AX,SCREEN_SEG    { Point ES:DI to the first rectangle pixel's    }
    MOV    ES,AX            {  address                                      }
    MOV    DX,SC_INDEX      { Set the Sequence Controller Index to point to }
    MOV    AL,MAP_MASK      {  the Map Mask register                        }
    OUT    DX,AL
    INC    DX               { Point DX to the SC Data register              }
    MOV    SI,WORD PTR StartX
    AND    SI,0003h                           { Look up left edge plane mask}
    MOV    BH,BYTE PTR LeftClipPlaneMask[SI]  {  to clip and put in BH      }
    MOV    SI,WORD PTR EndX
    AND    SI,0003h                           { Look up right edge plane    }
    MOV    BL,BYTE PTR RightClipPlaneMask[SI] {  mask to clip and put in BL }

    MOV    CX,WORD PTR EndX { Calculate number of addresses across rect     }
    MOV    SI,WORD PTR StartX
    CMP    CX,SI
    JLE    @FillDone        { Skip if 0 or negative width                   }
    DEC    CX
    AND    SI,NOT 011b
    SUB    CX,SI
    SHR    CX,1
    SHR    CX,1             { # of addresses across rectangle to fill - 1   }
    JNZ    @MasksSet        { There's more than one byte to draw            }
    AND    BH,BL            { There's only one byte, so combine the left    }
                            {  and right edge clip masks                    }
@MasksSet:
    MOV    SI,WORD PTR EndY
    SUB    SI,WORD PTR StartY { BX = height of rectangle                    }
    JLE    @FillDone        { Skip if 0 or negative height                  }
    MOV    AH,BYTE PTR Color { Color with which to fill                     }

    PUSH   BP

    MOV    BP,_SCREEN_WIDTH  { Stack frame isn't needed anymore              }
    SUB    BP,CX            { Distance from end of one scan line to start   }
    DEC    BP               {  of next                                      }
@FillRowsLoop:
    PUSH   CX               { Remember width in addresses - 1               }
    MOV    AL,BH            { Put left-edge clip mask in AL                 }
    OUT    DX,AL            { Set the left-edge plane (clip) mask           }
    MOV    AL,AH            { Put color in AL                               }
    STOSB                   { Draw the left edge                            }
    DEC    CX               { Count off left edge byte                      }
    JS     @FillLoopBottom  { That's the only byte                          }
    JZ     @DoRightEdge     { There are only two bytes                      }
    MOV    AL,00Fh          { Middle addresses are drawn 4 pixels at a pop  }
    OUT    DX,AL            { Set the middle pixel mask to no clip          }
    MOV    AL,AH            { Put color in AL                               }
    REP    STOSB            { Draw the middle addresses four pixels apiece  }
@DoRightEdge:
    MOV    AL,BL            { Put right-edge clip mask in AL                }
    OUT    DX,AL            { Set the right-edge plane (clip) mask          }
    MOV    AL,AH            { Put color in AL                               }
    STOSB                   { Draw the right edge                           }
@FillLoopBottom:
    ADD    DI,BP            { Point to the start of the next scan line of   }
                            {  the rectangle                                }
    POP    CX               { Retrieve width in addresses - 1               }
    DEC    SI               { Count down scan lines                         }
    JNZ    @FillRowsLoop

    POP    BP

@FillDone:
    POP    DI               { Restore caller's register variables           }
    POP    SI
    STI
  End; { Asm }
End; { FillRectangleX }

Procedure DrawLineX(X0,Y0,X1,Y1: Integer; PageBase: Word; Color: Word);
{ ------------------------------------------------------------------------- }
{ Mode X (320x240, 256 colors) line draw routine.  Fast assembler           }
{ implementation of Bresenham's line-drawing algorithm.                     }
{ ------------------------------------------------------------------------- }
Begin
  Asm
        MOV     DX,SC_INDEX     { Point to SC Index register                }
        MOV     AL,MAP_MASK
        OUT     DX,AL           { Point SC Index reg to the Map Mask        }
        PUSH    DS
        MOV     AX,SCREEN_SEG
        MOV     DS,AX
        MOV     SI,WORD PTR Y1 { Line Y start                          }
        MOV     AX,WORD PTR Y0 { Line Y end, used later in calculating }
                               { the start address                     }
        SUB     SI,AX          { Calculate DeltaY                      }
        JNS     @CalcStartAddress

{ DeltaY is negative -- swap coordinates so we're always working with a }
{ positive DeltaY.                                                      }

        MOV     AX,WORD PTR Y1 { Set line start to Y1, for use in }
                               { calculating the start address    }
        MOV     DX,WORD PTR X0
        XCHG    DX,WORD PTR X1
        MOV     WORD PTR X0,DX { Swap X coordinates               }
        NEG     SI             { Convert to positive DeltaY       }

{ Calculate the starting address in display memory of the line. }

@CalcStartAddress:
        MOV     CX,_SCREEN_WIDTH
        MUL     CX
        MOV     DI,AX
        MOV     CX,WORD PTR X0
        MOV     AX,CX
        SHR     CX,1
        SHR     CX,1
        ADD     DI,CX
        ADD     DI,WORD PTR PageBase

        AND     AX,3
        MOV     CX,AX
        MOV     AX,11h
        ROL     AL,CL
        MOV     DX,SC_INDEX+1
        OUT     DX,AL
        MOV     CX,AX
        MOV     AX,WORD PTR Color
        MOV     AH,AL
        MOV     AL,CL

        MOV     BX,WORD PTR X1 { Calculate DeltaX. }
        SUB     BX,WORD PTR X0

        PUSH    BP

        JS      @NegDeltaX { Handle correct one of four octants. }
        CMP     BX,SI
        JB      @Octant1

{ DeltaX >= DeltaY >= 0. }

{ ------------------------------------------------------------------------- }
{ LINE1 0                                                                   }
{ ------------------------------------------------------------------------- }
        MOV     CX,BX         { # of pixels in line                      }
        JCXZ    @Line1_End1    { Done if there are no more pixels         }
                              {  (there's always at least one pixel      }
                              {  at the start location)                  }
        SHL     SI,1          { DeltaY * 2                               }
        MOV     BP,SI         { Error term                               }
        SUB     BP,BX         { Error term starts at DeltaY * 2 - DeltaX }
        SHL     BX,1          { DeltaX * 2                               }
        SUB     SI,BX         { DeltaY * 2 - DeltaX * 2 (used in loop)   }
        ADD     BX,SI         { DeltaY * 2 (used in loop)                }
@LineLoop1:

{ See if it's time to advance the Y coordinate yet. }

        AND     BP,BP         { See if error term is negative      }
        JS      @MoveXCoord1  { Yes, stay at the same Y coordinate }

{ Advance the Y coordinate, first writing all pixels in the current byte. }

        MOV     BYTE PTR [DI],AH
        ADD     DI,_SCREEN_WIDTH
        ADD     BP,SI         { Adjust error term back down }
        JMP     @MoveToNextByte1

{ Adjust display memory address. }

@MoveXCoord1:
        ADD     BP,BX         { Increment error term & keep same Y coord }
        MOV     BYTE PTR [DI],AH
@MoveToNextByte1:
        ROL     AL,1          { Next pixel is in byte to right }
        ADC     DI,0
        OUT     DX,AL
        LOOP    @LineLoop1
@Line1_End1:
        MOV     BYTE PTR [DI],AH
        JMP     @VGALineDone

{ DeltaY > DeltaX >= 0. }

@Octant1:
{ -------------------------------------------------------------------- }
{ LINE2 0                                                              }
{ -------------------------------------------------------------------- }
{ Macro to loop through length of line, drawing each pixel in turn.    }
{ Used for case of DeltaX < DeltaY.                                    }
{ |DeltaY|+1 points are drawn.                                         }
{                                                                      }
{ Input:                                                               }
{       MOVE_LEFT: 1 if DeltaX < 0, 0 else                             }
{       AL: Line color                                                 }
{       BX: |DeltaX| (X distance between start and end points)         }
{       SI: DeltaY   (Y distance between start and end points)         }
{       ES:DI: Display-memory address of byte containing initial pixel }
{                                                                      }
{ Output: none                                                         }

        MOV     CX,SI         { # of pixels in line                      }
        SHL     BX,1          { DeltaX * 2                               }
        MOV     BP,BX         { Error term                               }
        SUB     BP,SI         { Error term starts at DeltaX * 2 - DeltaY }
        SHL     SI,1          { DeltaY * 2                               }
        SUB     BX,SI         { DeltaX * 2 - DeltaY * 2 (used in loop)   }
        ADD     SI,BX         { DeltaX * 2 (used in loop)                }

{ Draw initial pixel. }

        MOV     BYTE PTR [DI],AH
        JCXZ    @Line2_End3
@LineLoop3:

{ See if it's time to advance the X coordinate yet. }

        AND     BP,BP         { See if error term is negative            }
        JNS     @ETermAction3 { No, advance X coordinate                 }
        ADD     BP,SI         { Increment error term & keep same X coord }
        JMP     @MoveYCoord3
@ETermAction3:

{ Adjust display memory address. }

        ROL     AL,1
        ADC     DI,0
        OUT     DX,AL
        ADD     BP,BX         { Adjust error term back down }

{ Advance the Y coordinate. }

@MoveYCoord3:
        ADD     DI,_SCREEN_WIDTH

{ Write the next pixel. }

        MOV     BYTE PTR [DI],AH
        LOOP    @LineLoop3
@Line2_End3:
        JMP     @VGALineDone

@NegDeltaX:
        NEG     BX            { |DeltaX| }
        CMP     BX,SI
        JB      @Octant2

{ |DeltaX| >= DeltaY and DeltaX < 0. }

{ ------------------------------------------------------------------------- }
{ LINE1 1                                                                   }
{ ------------------------------------------------------------------------- }
        MOV     CX,BX         { # of pixels in line                      }
        JCXZ    @Line1_End2   { Done if there are no more pixels         }
                              {  (there's always at least one pixel      }
                              {  at the start location)                  }
        SHL     SI,1          { DeltaY * 2                               }
        MOV     BP,SI         { Error term                               }
        SUB     BP,BX         { Error term starts at DeltaY * 2 - DeltaX }
        SHL     BX,1          { DeltaX * 2                               }
        SUB     SI,BX         { DeltaY * 2 - DeltaX * 2 (used in loop)   }
        ADD     BX,SI         { DeltaY * 2 (used in loop)                }
@LineLoop2:

{ See if it's time to advance the Y coordinate yet. }

        AND     BP,BP         { See if error term is negative      }
        JS      @MoveXCoord2  { Yes, stay at the same Y coordinate }

{ Advance the Y coordinate, first writing all pixels in the current byte. }

        MOV     BYTE PTR [DI],AH
        ADD     DI,_SCREEN_WIDTH
        ADD     BP,SI         { Adjust error term back down }
        JMP     @MoveToNextByte2

{ Adjust display memory address. }

@MoveXCoord2:
        ADD     BP,BX         { Increment error term & keep same Y coord }
        MOV     BYTE PTR [DI],AH
@MoveToNextByte2:
        ROR     AL,1
        SBB     DI,0          { Next pixel is in byte to left  }
        OUT     DX,AL
        LOOP    @LineLoop2
@Line1_End2:
        MOV     BYTE PTR [DI],AH
        JMP     @VGALineDone

{ |DeltaX| < DeltaY and DeltaX < 0. }

@Octant2:
{ -------------------------------------------------------------------- }
{ LINE2 1                                                              }
{ -------------------------------------------------------------------- }
{ Macro to loop through length of line, drawing each pixel in turn.    }
{ Used for case of DeltaX < DeltaY.                                    }
{ |DeltaY|+1 points are drawn.                                         }
{                                                                      }
{ Input:                                                               }
{       MOVE_LEFT: 1 if DeltaX < 0, 0 else                             }
{       AL: Line color                                                 }
{       BX: |DeltaX| (X distance between start and end points)         }
{       SI: DeltaY   (Y distance between start and end points)         }
{       ES:DI: Display-memory address of byte containing initial pixel }
{                                                                      }
{ Output: none                                                         }

        MOV     CX,SI         { # of pixels in line                      }
        SHL     BX,1          { DeltaX * 2                               }
        MOV     BP,BX         { Error term                               }
        SUB     BP,SI         { Error term starts at DeltaX * 2 - DeltaY }
        SHL     SI,1          { DeltaY * 2                               }
        SUB     BX,SI         { DeltaX * 2 - DeltaY * 2 (used in loop)   }
        ADD     SI,BX         { DeltaX * 2 (used in loop)                }

{ Draw initial pixel. }

        MOV     BYTE PTR [DI],AH
        JCXZ    @Line2_End4
@LineLoop4:

{ See if it's time to advance the X coordinate yet. }

        AND     BP,BP         { See if error term is negative            }
        JNS     @ETermAction4 { No, advance X coordinate                 }
        ADD     BP,SI         { Increment error term & keep same X coord }
        JMP     @MoveYCoord4
@ETermAction4:

{ Adjust display memory address. }

        ROR     AL,1
        SBB     DI,0
        OUT     DX,AL
        ADD     BP,BX         { Adjust error term back down }

{ Advance the Y coordinate. }

@MoveYCoord4:
        ADD     DI,_SCREEN_WIDTH

{ Write the next pixel. }

        MOV     BYTE PTR [DI],AH
        LOOP    @LineLoop4
@Line2_End4:
@VGALineDone:

{ Pop registers and exit. }

        POP     BP
        POP     DS
  End; { Asm }
End; { DrawLineX }


Procedure RectangleX(StartX,StartY,EndX,EndY: Integer; PageBase: Word; Color: Word);
{ ------------------------------------------------------------------------- }
{ Mode X (320x240, 256 colors) rectangle plot routine.                      }
{ ------------------------------------------------------------------------- }
Begin
  DrawLineX(StartX,StartY,EndX  ,StartY,PageBase,Color);
  DrawLineX(EndX  ,StartY,EndX  ,EndY  ,PageBase,Color);
  DrawLineX(EndX  ,EndY  ,StartX,EndY  ,PageBase,Color);
  DrawLineX(StartX,EndY  ,StartX,StartY,PageBase,Color);
End; { RectangleX }

Procedure CopySystemToScreenX(SourceStartX,SourceStartY,SourceEndX,SourceEndY,
  DestStartX,DestStartY: Integer; SourcePtr: Pointer; DestPageBase: Word;
  SourceBitmapWidth,DestBitmapWidth: Integer);
{ ------------------------------------------------------------------------- }
{ Mode X (320x240,256 colors) system memory to display memory copy routine. }
{ Uses approach of changing the plane for each pixel copied; this is slower }
{ than copying all pixels in one plane, then all pixels in the next plane,  }
{ and so on, but it is simpler; besides, images for which performance is    }
{ critical should be stored in off-screen memory and copied to the screen   }
{ via the latches. Copies up to but not including the column at SourceEndX  }
{ and the row at SourceEndY. No clipping is performed.                      }
{ Pascal near-callable as:                                                  }
{    CopySystemToScreenX(SourceStartX, SourceStartY,SourceEndX,SourceEndY,  }
{       DestStartX,DestStartY,SourcePtr,DestPageBase,SourceBitmapWidth,     }
{       DestBitmapWidth);                                                   }
{ ------------------------------------------------------------------------- }

Var
  RectWidth : Word;         { Local storage for width of rectangle          }
  LeftMask  : Byte;         { Local storage for left rect edge plane mask   }
  Counter0  : Word;
  Counter   : Word;

Begin
  Asm
    CLI
    PUSH    DS
    PUSH    SI              { Preserve caller's register variables          }
    PUSH    DI

    CLD
    MOV     AX,SCREEN_SEG   { Point ES to display memory                    }
    MOV     ES,AX
    MOV     AX,WORD PTR SourceBitmapWidth
    MUL     WORD PTR SourceStartY { Top source rect scan line               }
    ADD     AX,WORD PTR SourceStartX

    LDS     SI,DWORD PTR SourcePtr { Offset of first source rect pixel in DS}
    ADD     SI,AX

    MOV     AX,WORD PTR DestBitmapWidth
    SHR     AX,2            { Convert to width in addresses                 }
    MOV     WORD PTR DestBitmapWidth,AX  { Remember address width           }
    MUL     WORD PTR DestStartY { Top dest rect scan line                   }
    MOV     DI,WORD PTR DestStartX
    MOV     CX,DI
    SHR     DI,2          { X/4=offset of first dest rect pixel in scan line}
    ADD     DI,AX           { offset of first dest rect pixel in page       }
    ADD     DI,WORD PTR DestPageBase { Offset of first dest rect pixel in   }
                                     {  display memory                      }
    AND     CL,011b         { CL=first dest pixel's plane                   }
    MOV     AL,11h          { upper nibble comes into play when plane wraps }
                            {  from 3 back to 0                             }
    SHL     AL,CL                { Set the bit for the first dest pixel's   }
    MOV     BYTE PTR LeftMask,AL {  plane in each nibble to 1               }

    MOV     CX,WORD PTR SourceEndX  { Calculate # of pixels across rect     }
    SUB     CX,WORD PTR SourceStartX
    JLE     @CopyDone       { Skip if 0 or negative width                   }
    MOV     WORD PTR RectWidth,CX
    MOV     BX,WORD PTR SourceEndY
    SUB     BX,WORD PTR SourceStartY  { BX=height of rectangle              }
    JLE     @CopyDone       { Skip if 0 or negative height                  }
    MOV     WORD PTR Counter,BX
    MOV     DX,SC_INDEX     { Point to SC Index register                    }
    MOV     AL,MAP_MASK
    OUT     DX,AL           { Point SC Index reg to the Map Mask            }
    INC     DX              { Point DX to SC Data reg                       }

    MOV     AX,WORD PTR Counter
    MOV     WORD PTR Counter0,AX
    PUSH    SI              { Remember the start offset in the source       }
    PUSH    DI              { Remember the start offset in the dest         }
    MOV     AL,BYTE PTR LeftMask
    OUT     DX,AL
@CopyRowsLoop1:
    MOV     CX,WORD PTR RectWidth
    PUSH    SI              { Remember the start offset in the source       }
    PUSH    DI              { Remember the start offset in the dest         }
    ADD     CX,3
    SHR     CX,2
    TEST    CX,3
    JZ      @Plane1A
    MOV     AX,CX
@Plane1C:
    AND     CX,3
    MOVSB
    ADD     SI,3
    DEC     CX
    JNZ     @Plane1C
    MOV     CX,AX
@Plane1A:
    SHR     CX,2
    JZ      @Plane1B
@Plane1:
    MOV     AL,BYTE PTR [SI + 8]
    MOV     AH,BYTE PTR [SI + 12]
    DB      _32BIT
    SHL     AX,16
    MOV     AL,BYTE PTR [SI]
    MOV     AH,BYTE PTR [SI + 4]
    DB      _32BIT
    STOSW
    ADD     SI,16
    DEC     CX
    JNZ     @Plane1         { This is actually faster than a LOOP on a 386  }
@Plane1B:
    POP     DI              { Retrieve the dest start offset                }
    ADD     DI,WORD PTR DestBitmapWidth { Point to the start of the next    }
                                        {  scan line of the dest            }
    POP     SI              { Retrieve the source start offset              }
    ADD     SI,WORD PTR SourceBitmapWidth { Point to the start of the next  }
                                          {  scan line of the source        }
    DEC     WORD PTR Counter  { Count down scan lines                       }
    JNZ     @CopyRowsLoop1
    MOV     AX,WORD PTR Counter0
    MOV     WORD PTR Counter,AX

    POP     DI
    POP     SI
    INC     SI
    MOV     AL,BYTE PTR LeftMask
    ROL     AL,1
    ADC     DI,0
    MOV     BYTE PTR LeftMask,AL
    OUT     DX,AL
    PUSH    SI              { Remember the start offset in the source       }
    PUSH    DI              { Remember the start offset in the dest         }
@CopyRowsLoop2:
    PUSH    SI
    PUSH    DI
    MOV     CX,WORD PTR RectWidth
    ADD     CX,2
    SHR     CX,2
    TEST    CX,3
    JZ      @Plane2A
    MOV     AX,CX
@Plane2C:
    AND     CX,3
    MOVSB
    ADD     SI,3
    DEC     CX
    JNZ     @Plane2C
    MOV     CX,AX
@Plane2A:
    SHR     CX,2
    JZ      @Plane2B
@Plane2:
    MOV     AL,BYTE PTR [SI + 8]
    MOV     AH,BYTE PTR [SI + 12]
    DB      _32BIT
    SHL     AX,16
    MOV     AL,BYTE PTR [SI]
    MOV     AH,BYTE PTR [SI + 4]
    DB      _32BIT
    STOSW
    ADD     SI,16
    DEC     CX
    JNZ     @Plane2         { This is actually faster than a LOOP on a 386  }
@Plane2B:
    POP     DI              { Retrieve the dest start offset                }
    ADD     DI,WORD PTR DestBitmapWidth { Point to the start of the next    }
                                        {  scan line of the dest            }
    POP     SI              { Retrieve the source start offset              }
    ADD     SI,WORD PTR SourceBitmapWidth { Point to the start of the next  }
                                          {  scan line of the source        }
    DEC     WORD PTR Counter  { Count down scan lines                       }
    JNZ     @CopyRowsLoop2
    MOV     AX,WORD PTR Counter0
    MOV     WORD PTR Counter,AX

    POP     DI
    POP     SI
    INC     SI
    MOV     AL,BYTE PTR LeftMask
    ROL     AL,1
    ADC     DI,0
    MOV     BYTE PTR LeftMask,AL
    OUT     DX,AL
    PUSH    SI              { Remember the start offset in the source       }
    PUSH    DI              { Remember the start offset in the dest         }
@CopyRowsLoop3:
    PUSH    SI
    PUSH    DI
    MOV     CX,WORD PTR RectWidth
    INC     CX
    SHR     CX,2
    TEST    CX,3
    JZ      @Plane3A
    MOV     AX,CX
@Plane3C:
    AND     CX,3
    MOVSB
    ADD     SI,3
    DEC     CX
    JNZ     @Plane3C
    MOV     CX,AX
@Plane3A:
    SHR     CX,2
    JZ      @Plane3B
@Plane3:
    MOV     AL,BYTE PTR [SI + 8]
    MOV     AH,BYTE PTR [SI + 12]
    DB      _32BIT
    SHL     AX,16
    MOV     AL,BYTE PTR [SI]
    MOV     AH,BYTE PTR [SI + 4]
    DB      _32BIT
    STOSW
    ADD     SI,16
    DEC     CX
    JNZ     @Plane3         { This is actually faster than a LOOP on a 386  }
@Plane3B:
    POP     DI              { Retrieve the dest start offset                }
    ADD     DI,WORD PTR DestBitmapWidth { Point to the start of the next    }
                                        {  scan line of the dest            }
    POP     SI              { Retrieve the source start offset              }
    ADD     SI,WORD PTR SourceBitmapWidth { Point to the start of the next  }
                                          {  scan line of the source        }
    DEC     WORD PTR Counter  { Count down scan lines                       }
    JNZ     @CopyRowsLoop3
    MOV     AX,WORD PTR Counter0
    MOV     WORD PTR Counter,AX

    POP     DI
    POP     SI
    INC     SI
    MOV     AL,BYTE PTR LeftMask
    ROL     AL,1
    ADC     DI,0
    OUT     DX,AL
    PUSH    SI              { Remember the start offset in the source       }
    PUSH    DI              { Remember the start offset in the dest         }
@CopyRowsLoop4:
    PUSH    SI
    PUSH    DI
    MOV     CX,WORD PTR RectWidth
    SHR     CX,2
    TEST    CX,3
    JZ      @Plane4A
    MOV     AX,CX
@Plane4C:
    AND     CX,3
    MOVSB
    ADD     SI,3
    DEC     CX
    JNZ     @Plane4C
    MOV     CX,AX
@Plane4A:
    SHR     CX,2
    JZ      @Plane4B
@Plane4:
    MOV     AL,BYTE PTR [SI + 8]
    MOV     AH,BYTE PTR [SI + 12]
    DB      _32BIT
    SHL     AX,16
    MOV     AL,BYTE PTR [SI]
    MOV     AH,BYTE PTR [SI + 4]
    DB      _32BIT
    STOSW
    ADD     SI,16
    DEC     CX              { This is actually faster than a LOOP on a 386  }
    JNZ     @Plane4
@Plane4B:
    POP     DI              { Retrieve the dest start offset                }
    ADD     DI,WORD PTR DestBitmapWidth { Point to the start of the next    }
                                        {  scan line of the dest            }
    POP     SI              { Retrieve the source start offset              }
    ADD     SI,WORD PTR SourceBitmapWidth { Point to the start of the next  }
                                          {  scan line of the source        }
    DEC     WORD PTR Counter  { Count down scan lines                       }
    JNZ     @CopyRowsLoop4

    POP     DI
    POP     SI
(*
@CopyScanLineLoop:
    OUT     DX,AL           { Set the plane for this pixel                  }
    MOVSB                   { Copy the pixel to the screen                  }
    ROL     AL,1            { Set mask for next pixel's plane               }
    CMC                     { advance destination address only when wrapping}
    SBB     DI,0            {  from plane 3 to plane 0                      }
                            {  (else undo INC DI done by MOVSB)             }
    LOOP    @CopyScanLineLoop
    POP     DI              { Retrieve the dest start offset                }
    ADD     DI,WORD PTR DestBitmapWidth { Point to the start of the next    }
                                        {  scan line of the dest            }
    POP     SI              { Retrieve the source start offset              }
    ADD     SI,WORD PTR SourceBitmapWidth { Point to the start of the next  }
                                          {  scan line of the source        }
    DEC     WORD PTR Counter  { Count down scan lines                       }
    JNZ     @CopyRowsLoop
*)
@CopyDone:
    POP     DI              { Restore caller's register variables           }
    POP     SI
    POP     DS
    STI
  End; { Asm }
End; { CopySystemToScreenX }

End.
